from tqdm import tqdm
import fsspec
import os
from joblib import Parallel, delayed, cpu_count
from zipfile import ZipFile

from lost_ds.functional.mapping import remap_img_path
from lost_ds.util import get_fs


def copy_imgs(df, out_dir, col='img_path', force_overwrite=False, 
              filesystem=None, parallel=-1):
    '''Copy all images of dataset into out_dir

    Args:
        df (pd.DataFrame): dataframe to copy
        out_dir (str): Destination folder to store images
        col (str): column containing paths to files
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
    '''
    fs = get_fs(filesystem)
    def copy_file(src_path, dst_dir):
        dst_path = os.path.join(dst_dir, os.path.basename(src_path))
        if fs.exists(dst_path) and not force_overwrite:
            return
        fs.copy(src_path, dst_path)
        
    fs.makedirs(out_dir, exist_ok=True)
    img_paths = list(df[col].unique())
    if parallel:
        Parallel(n_jobs=parallel)(delayed(copy_file)(path, out_dir) 
                            for path in tqdm(img_paths, desc='copy imgs'))
    else:
        for path in tqdm(img_paths, desc='copy imgs'):
            copy_file(path, out_dir)
        

def copy_to_zip(zip_file, df, zip_dir, col='img_path', 
              filesystem=None, progress_callback=None):
    '''Copy all images of dataset into zip archive

    Args:
        df (pd.DataFrame): dataframe to copy
        zip_root (str): Root path in zip archive
        col (str): column containing paths to files
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        progress_callback (function): Will be called on progress
            callback definition -> progress_callback(progress), where progress 
            value is between 0...100 
    '''

    fs = get_fs(filesystem).fs
    def copy_file_to_zip(src_path, zip_file):
        dst_path = os.path.join(zip_dir, os.path.basename(src_path))
        try:
            fs.ls('')
        except:
            pass
        with fs.open(src_path, 'rb') as f:
            zip_file.writestr(dst_path, f.read())
        
    img_paths = list(df[col].unique())
    total = len(img_paths)
    next_pg = 0
    for idx, path in enumerate(img_paths):
        copy_file_to_zip(path, zip_file)
        if progress_callback is not None:
            pg = (idx+1) *100 / total
            if pg == 100:
                progress_callback(pg)
            elif pg >= next_pg:
                progress_callback(pg)
                next_pg += 5
    
def pack_ds(df, out_dir, cols=['img_path', 'mask_path', 'crop_path'],
            dirs = ['imgs', 'masks', 'crops'], filesystem=None, zip_file=None, 
            progress_callback=None):
    '''Copy all images from dataset to a new place and update the dataframe 
    
    Args:
        df (pd.DataFrame): Dataframe to copy
        out_dir (str): Name of the directory to store the information
        executor (Client, ThreadPoolExecutor): executor for parallelization
            if None a new ThreadPoolExecutor will be initialized
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
        cols (list of string): column names containing file-paths
        dirs (list of string): name of new directories according to cols. The 
            dirs will contain the copied data
        zip_file (zipfile.ZipFile or None): If not None, a ZipFile object will 
            be used to pack dataset to zip archive  
        progress_callback (function): Will be called on progress
            callback definition -> progress_callback(progress), where progress 
            value is between 0...100 
        
    Returns:
        pd.DataFrame with new image paths
    '''
    fs = get_fs(filesystem)
    for col, _dir in zip(cols, dirs):
        if col in df.keys():
            dout = os.path.join(out_dir, _dir)
            if zip_file is None:
                if progress_callback is not None:
                    raise Exception('progress_callback is only implement for packing to zip files yet!')
                fs.makedirs(dout, exist_ok=True)
                copy_imgs(df=df, out_dir=dout, col=col, filesystem=fs)
            else:
                out_base = os.path.basename(out_dir)
                out_base = os.path.splitext(out_base)[0]
                zip_dir = os.path.join(out_base, _dir)
                copy_to_zip(zip_file, df, zip_dir=zip_dir, col=col, filesystem=fs, 
                            progress_callback=progress_callback)
            df = remap_img_path(df, dout, col)
    return df


from lost_ds.functional.split import (split_by_empty,
                                     split_by_img_path,
                                     split_multilabels,
                                     split_train_test
                                     )

from lost_ds.functional.filter import (remove_empty,
                                      img_selection,
                                      label_selection,
                                      unique_labels,
                                      ignore_labels,
                                      selection_mask,
                                      is_multilabel,                                   
                                      )

from lost_ds.functional.validation import (validate_empty_images,
                                          validate_unique_annos,
                                          validate_geometries,
                                          validate_img_paths, 
                                          validate_single_labels
                                          )

from lost_ds.functional.transform import (to_abs, 
                                          to_rel,
                                          transform_bbox_style,
                                          polygon_to_bbox,
                                          to_coco)

from lost_ds.functional.mapping import (remap_img_path,
                                        remap_labels)


import os

from tqdm import tqdm
from joblib import Parallel, delayed, cpu_count
try:
    import seaborn as sns
except:
    pass
import numpy as np 
import cv2 

from lost_ds.util import get_fs
from lost_ds.geometry.lost_geom import LOSTGeometries
from lost_ds.functional.api import remove_empty


def get_fontscale(fontscale, thickness, img_h, text_max_h_frac=0.04):
    if isinstance(fontscale, (int, float)):
        return fontscale
    elif fontscale=='auto':
        text_h = int(text_max_h_frac * img_h)
        fontscale = cv2.getFontScaleFromHeight(cv2.FONT_HERSHEY_SIMPLEX, 
                                           max(text_h, 10), 
                                           thickness)
    return fontscale


def get_thickness(line_thickness, img_h, thickness_max_h_frac=0.002):
    if line_thickness == 'auto':
        return int(thickness_max_h_frac * img_h)
    else: 
        return line_thickness


def vis_sample(img, df, line_thickness=3, color=(0, 0, 255), 
               lbl_col='anno_lbl', lost_geometries:LOSTGeometries=None, 
               blow_up=None, radius=2, fontscale=2):
    '''Visualize annos of an image

    Args:
        img (np.ndarray): image to draw on
        df (pandas.DataFrame): The DataFrame that contains annoations to 
            visualize. If df is None a random image from df will be 
            sampled.
        color (tuple, dict of tuple, string): colors (B,G,R) for all annos 
            if tuple, dict for labelwise mapping like {label: color}, string for 
            column in dataframe containing the color
        line_thickness (int, dict of int): line thickness for annotations if int
            or dict for anno-type wise mapping like {dtype: thickness}
        lost_geometries (LOSTGeometries): LOSTGeometries instance to use, will
            create a new one if None
        blow_up (): TODO: implement
    Returns:
        np.array: Image painted with annotations.
    '''
    df = remove_empty(df, 'anno_data')
    if len(df) > 0:
        geom = lost_geometries
        if lost_geometries is None:
            geom = LOSTGeometries()
        
        anno_data = list(df['anno_data'])
        anno_conf = None
        if hasattr(df, 'anno_confidence'): 
            anno_conf = list(df['anno_confidence']) 
        if lbl_col is None:
            anno_lbl = [None] * len(anno_data)
        else:
            anno_lbl = list(df[lbl_col])
        anno_dtype = list(df['anno_dtype'])
        anno_style = list(df['anno_style'])
        anno_format = list(df['anno_format'])
        if isinstance(color, str):
            color = list(df[color])
        thickness = get_thickness(line_thickness, img.shape[0])
        fontscale = get_fontscale(fontscale, thickness, img.shape[0])            
        thickness = max(1, thickness)
        img = geom.draw(img, anno_data, anno_conf, anno_lbl, anno_dtype, 
                        anno_style, anno_format, thickness, fontscale, color, 
                        radius)
            
    return img


def vis_and_store(df, out_dir, lbl_col='anno_lbl', color=(0, 0, 255), 
                  line_thickness=2, fontscale=2, filesystem=None, 
                  radius=2, parallel=-1):
    '''Visualize annotations and store them to a folder

    Args:
        df (pd.DataFrame): Optional dataset in lost format to visualize
        out_dir (str): Directory to store the visualized annotations
        color (tuple, dict of tuple, string): colors (B,G,R) for all annos 
            if tuple, dict for labelwise mapping like {label: color}, string for 
            column in dataframe containing the color
        line_thickness (int, dict of int): line thickness for annotations if int
            or dict for anno-type wise mapping like {dtype: thickness}
        lbl_col (str, None): column containing the labels, draws no lbl if None
        radius (int): radius to draw for points/circles
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
    '''
    fs = get_fs(filesystem)
    fs.makedirs(out_dir, exist_ok=True)
    
    def vis_img(img_path, df_vis):
        geom = LOSTGeometries()
        out_path = os.path.join(out_dir, os.path.basename(img_path))
        if df_vis['anno_data'].notnull().any():
            img = fs.read_img(img_path)
            img = vis_sample(img=img, df=df_vis, line_thickness=line_thickness, 
                             color=color, lbl_col=lbl_col, lost_geometries=geom, 
                             radius=radius, fontscale=fontscale)
            fs.write_img(img, out_path)
        else:
            fs.copy(img_path, out_path)
    if parallel:
        Parallel(n_jobs=parallel)(delayed(vis_img)(path, df_vis) 
                                  for path, df_vis in tqdm(df.groupby('img_path'), 
                                                           desc='visualize'))
    else:
        for path, df_vis in tqdm(df.groupby('img_path'), desc='visualize'):
            vis_img(path, df_vis) 
    

def vis_semantic_segmentation(df, out_dir, n_classes, palette='dark', 
                              seg_path_col='seg_path', filesystem=None, 
                              parallel=-1):
    """Visualize the stored semantic segmentations by coloring it
    
    Args:
        df (pandas.DataFrame): The DataFrame that contains annoations to 
            visualize. 
        out_dir (str): path to store images
        n_classes (int): number of classes occuring in pixelmaps, number of 
            different colors needed for visualization
        palette (str): seaborn color palette i.e. 'dark', 'bright', 'pastel',...
            refer https://seaborn.pydata.org/tutorial/color_palettes.html 
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
    """
    fs = get_fs(filesystem)
    fs.makedirs(out_dir, exist_ok=True)
    
    palette = sns.color_palette(palette, n_classes)
    palette = [(np.array(x)*255).astype(np.uint8) for x in palette]

    def vis_seg(seg_path):
        seg = fs.read_img(seg_path)
        vis = np.zeros(seg.shape[:2] + (3,))
        for i in range(n_classes):
            vis = np.where(seg==i, palette[i], vis)
        fs.write_img(vis, os.path.join(out_dir, seg_path.split('/')[-1]))
    
    seg_paths = df[seg_path_col].unique()
    
    if parallel:
        Parallel(n_jobs=parallel)(delayed(vis_seg)(seg_path) 
                for seg_path in tqdm(seg_paths, desc='vis sem. seg.'))
    else:
        for seg_path in tqdm(seg_paths, desc='vis sem. seg.'):
            vis_seg(seg_path)



from ast import literal_eval
import json
import pandas as pd
from lost_ds.core import LOSTDataset


def old_lds_to_new_lds(df):
    # load data (unstring)
    def literal_eval_parse(entry):
        if not pd.isnull(entry):
            if isinstance(entry, str):
                return literal_eval(entry.replace('nan', 'None'))
            else:
                return literal_eval(entry)
        else:
            return None

    def parse_col(col):
        try:
            return col.map(lambda entry: literal_eval_parse(entry))
        except Exception as e:
            try:
                return col.map(lambda entry: json.loads(entry))
            except Exception as e:
                return col
            
    df = df.apply(lambda x: parse_col(x), axis=0)
    # map keys
    mapping = { 'anno.idx': 'anno_uid',
                'anno.timestamp': 'anno_timestamp',
                'anno.state': 'anno_state',
                'anno.dtype':'anno_dtype',
                'anno.sim_class':'anno_sim_class',
                'anno.iteration':'anno_iteration',
                'anno.user_id':'anno_user_id',
                'anno.user':'anno_user',
                'anno.confidence':'anno_confidence',
                'anno.anno_time':'anno_anno_time',
                'anno.data': 'anno_data',
                'anno.lbl.name':'anno_lbl',
                'img.idx': 'img_uid',
                'img.timestamp': 'img_timestamp',
                'img.state': 'img_state',
                'img.sim_class': 'img_sim_class',
                'img.frame_n': 'img_frame_n',
                'img.img_path': 'img_path',
                'img.iteration': 'img_iteration',
                'img.user_id': 'img_user_id',
                'img.anno_time': 'img_anno_time',
                'img.lbl.name': 'img_lbl',
                'img.annotator': 'img_user',
                'img.is_junk': 'img_is_junk'
                }
    df_map = df.rename(columns=mapping)
    new_keys = ['anno_style', 'anno_format']
    df_map[new_keys] = None

    # # transform timestamp data
    # all_keys = list(mapping.values())
    # for k in all_keys:
    #     if 'timestamp' in k:
    #         df_map[k] = df_map[k].map(lambda x: str(x))
    
    # transform anno data
    def parse_data(row):
        dtype = row['anno_dtype']
        data = None
        style = None
        frmt = 'rel'
        if dtype == 'bbox':
            data = list(row['anno_data'].values())
            style = 'xcycwh'
        elif dtype in ['polygon', 'line']:
            data = [list(p.values()) for p in row['anno_data']]
            style = 'xy'
        elif dtype == 'point':
            data = list(row['anno_data'].values())
            style = 'xy'
        return data, style, frmt
    new_dat = df_map.apply(lambda x: parse_data(x), axis=1)
    dic = {'anno_data': [],
           'anno_style': [],
           'anno_format': []}
    
    indexes = []
    for i, dat in new_dat.iteritems():
        indexes.append(i)
        dic['anno_data'].append(dat[0])
        dic['anno_style'].append(dat[1])
        dic['anno_format'].append(dat[2])
    df_dat = pd.DataFrame(dic, index=indexes)
    df_map[['anno_data', 'anno_style', 'anno_format']] = df_dat[['anno_data', 'anno_style', 'anno_format']]
    
    drop_keys = [k for k in list(df_map.keys()) if '.' in k]
    df_map.drop(labels=drop_keys, axis='columns', inplace=True)
    
    def _lbl_parser(lbl):
        if isinstance(lbl, str): return lbl.lower()
        else: return lbl
    df_map['anno_lbl'] = df_map['anno_lbl'].map(lambda x: [_lbl_parser(lbl) for lbl in x] if isinstance(x, list) else x)
    
    return LOSTDataset(df_map)




import numpy as np
from shapely.geometry import box, Polygon, MultiPolygon
import cv2

from lost_ds.geometry.api import Geometry
from lost_ds.vis.geometries import draw_boxes

class Bbox(Geometry):
    
    styles = {
              'lost':       'xcycwh',     # LOST-style
              'opencv':     'xywh',       # COCO and openCV-style
              'shapely':    'x1y1x2y2'    # Pascal VOC and shapely style
              }    
    
    def __init__(self):
        super().__init__()
    
        
    def from_polygon(self, data, dst_style):
        '''Create a bbox from a polygon
        Args:
            row (pd.Series): row with attributes 'anno_data', 'anno_style', 
                'anno_dtype'
            dst_style (str): dst style for bbox
        Returns:
            pd.Series like row with bbox data
        '''
        # bounds gets us xmin, ymin, xmax, ymax
        bbox = np.array(Polygon(data).bounds)
        anno_style = 'x1y1x2y2'
        if dst_style is None:
            dst_style = anno_style
        data = self.transform_style(bbox, anno_style, dst_style)
        return data
    
    
    def validate(self, data):
        '''Validate bbox data by shape
        Args:
            data (list): bbox data
        Returns:
            bool True for valid False for invalid
        '''
        return len(data.shape)==1 and len(data)==4
    
    
    def to_shapely(self, data, style):
        '''Get a shapely box Instance
        Args:
            data (list, np.ndarray): bbox data
            style (str): current bbox-style
        Returns: 
            shapely box instance 
        '''
        bbox = self.transform_style(data, style, self.styles['shapely'])
        return box(*bbox)
    
    
    def crop(self, crop_pos, data, style):
        intersection, intersects = self._crop_intersection(crop_pos, data, style=style)
        if not intersects:
            return intersection
        xmin, ymin, xmax, ymax = crop_pos.bounds
        new_bboxes = []
        if isinstance(intersection, MultiPolygon):
            new_bboxes = list(intersection.geoms)
        else:
            new_bboxes = [intersection]
            
        for i, bbox in enumerate(new_bboxes):
            new_bbox = np.array(bbox.bounds).squeeze()-[xmin, ymin, xmin, ymin]
            new_bbox = self.transform_style(new_bbox, self.styles['shapely'], 
                                            style)
            new_bboxes[i] = new_bbox
        return new_bboxes
    
    
    def _draw(self, img, data, style, text, color, line_thickness, fontscale, 
              **kwargs):
        if line_thickness is None:
            line_thickness = 2
        data = self.transform_style(data, style, 'x1y1x2y2')
        return draw_boxes(img, data, text, color, line_thickness, fontscale)
    
    
    def segmentation(self, segmentation, color, anno_data, anno_format, 
                     anno_style, **kwargs):
        anno_data = self.to_abs(anno_data, anno_format, segmentation.shape)
        anno_data = self.transform_style(anno_data, anno_style, 
                                        self.styles['shapely']).astype(np.int32)
        p1, p2 = anno_data[:2], anno_data[2:]
        cv2.rectangle(segmentation, tuple(p1), tuple(p2), color, cv2.FILLED)
        return segmentation
    
    
    #
    #   style transformation
    #
    
    def xcycwh_trans(self, data, style):
        '''Transform xcycwh style
        Args:
            data (np.ndarray, list): bbox data
            style (str): new bbox style, one of {xywh, x1y1x2y2}
        Returns:
            list with bbox in new style
        '''
        xc,yc,w,h = data
        x1, y1 = xc - w/2, yc - h/2
        if style == 'xywh': return [x1, y1, w, h]
        elif style == 'x1y1x2y2': return [x1, y1, x1 + w, y1 + h]
    
    
    def xywh_trans(self, data, style):
        '''Transform xywh style
        Args:
            data (np.ndarray, list): bbox data
            style (str): new bbox style, one of {xcycwh, x1y1x2y2}
        Returns:
            list with bbox in new style
        '''
        x,y,w,h = data
        if style == 'xcycwh': return [x + w/2, y + h/2, w, h]
        elif style == 'x1y1x2y2': return [x, y, x + w, y + h]
    
    
    def x1y1x2y2_trans(self, data, style):
        '''Transform x1y1x2y2 style
        Args:
            data (np.ndarray, list): bbox data
            style (str): new bbox style, one of {xcycwh, xywh}
        Returns:
            list with bbox in new style
        '''
        x1, y1, x2, y2 = data
        w, h = x2 - x1, y2 - y1
        if style == 'xcycwh': return [x1 + w/2, y1 + h/2, w, h]
        elif style == 'xywh': return [x1, y1, w, h]
            
            
    def transform_style(self, data, src_style, dst_style):
        '''transform style of bbox
        Args:
            data (numpy.ndarray): bbox
            src_style (str): current style: {'xcycwh', 'xywh', 'x1y1x2y2'}
            dst_style (str): desired style: {'xcycwh', 'xywh', 'x1y1x2y2'}
        Returns:
            numpy.ndarray: bbox in dst_style
        '''
        if src_style == dst_style:
            return data
        new_style_data = None
        if src_style == 'xywh': 
            new_style_data = self.xywh_trans(data, dst_style)
        elif src_style == 'x1y1x2y2':
            new_style_data = self.x1y1x2y2_trans(data, dst_style)
        elif src_style == 'xcycwh':
            new_style_data = self.xcycwh_trans(data, dst_style)
        if new_style_data is None:
            raise Exception('Unknown bbox-style or empty anno-data!')
        return np.array(new_style_data)
    


from shapely.geometry import Point as Pt, MultiPoint
import numpy as np
import cv2

from lost_ds.geometry.api import Geometry
from lost_ds.vis.geometries import draw_points


class Point(Geometry):
    
    def __init__(self):
        super().__init__()


    def to_shapely(self, data):
        return Pt(data)
    
    
    def segmentation(self, segmentation, color, anno_data, anno_format, 
                     anno_style, radius, **kwargs):
        anno_data = self.to_abs(anno_data, anno_format, segmentation.shape)
        if radius is None:
            return segmentation
        cv2.circle(segmentation, tuple(anno_data.astype(np.int32)), radius, 
                   color, cv2.FILLED)
        return segmentation
        
    
    def crop(self, crop_pos, data, **kwargs):
        intersection, intersects = self._crop_intersection(crop_pos, data)
        if not intersects:
            return intersection
        xmin, ymin, xmax, ymax = crop_pos.bounds
        new_points = []
        if isinstance(intersection, MultiPoint):
            new_points = list(intersection.geoms)
        else:
            new_points = [intersection]
            
        for i, point in enumerate(new_points):
            new_point = np.array(point.coords) - [xmin, ymin]
            new_points[i] = new_point.squeeze()
        
        return new_points
    

    def validate(self, data):
        return len(data.shape)==1 and len(data)==2

    
    def _draw(self, img, data, style, text, color, line_thickness, fontscale, 
              radius):
        if line_thickness is None:
            line_thickness = -1
        return draw_points(img, data, text, color, radius, line_thickness, 
                           fontscale)


import pandas as pd 
import pyarrow as pa
from joblib import parallel_backend, Parallel, delayed, effective_n_jobs, cpu_count
from sklearn.utils import gen_even_slices
from sklearn.utils.validation import _num_samples
# from lost_ds.core import LOSTDataset

from lost_ds.io.file_man import FileMan
from lost_ds.geometry.lost_geom import LOSTGeometries

   
def get_fs(filesystem, backend='pandas'):
    if isinstance(filesystem, FileMan):
        return filesystem
    else:
        return FileMan(filesystem, backend)


def prep_parquet(df, try_serialize_objects=False):
    geom = LOSTGeometries()
    store_df = df.copy()
    if 'anno_data' in df.keys():
        store_df.anno_data = store_df.anno_data.map(lambda x: 
            geom.serializable(x))
    
    # HACK: If data of different dimensions is provided pyarrow wont serialize without errors
    # try which object-columns cannot be serialized with pyarrow and try to give them
    # a 2 dimensional shape
    if try_serialize_objects:
        object_columns = (store_df.dtypes=='O').index[(store_df.dtypes=='O').values]
        for k in object_columns:
            try:
                pa.table(store_df[[k]])
            except pa.ArrowInvalid:
                store_df[k] = store_df[k].map(lambda x: geom.serializable(x))
                try:
                    pa.table(store_df[[k]])
                except:
                    raise Exception(pa.ArrowInvalid, f'Cannot serialize data of column {k}!')
                store_df.rename({k: k + '_lds_serialized'}, axis=1, inplace=True)
    return store_df


def to_parquet(path, df, filesystem=None):
    fs = get_fs(filesystem)
    store_df = prep_parquet(df)
    try:
        fs.write_dataset(store_df, path)
    except pa.ArrowInvalid:
        store_df = prep_parquet(df, try_serialize_objects=True)
        fs.write_dataset(store_df, path)
    
    
def parallel_apply(df:pd.DataFrame, func, n_jobs= -1, **kwargs):
    """ Pandas apply in parallel using joblib. 
    Uses sklearn.utils to partition input evenly.
    
    Args:
        df (pd.DataFrame, pd.Series): dataframe to apply the function 
        func (callable): Callable function to apply
        n_jobs (int): Desired number of workers. Default value -1 means use all 
            available cores.
        **kwargs: Any additional parameters will be supplied to the apply function
        
    Returns:
        Same as for normal Pandas DataFrame.apply()
    """
    
    if effective_n_jobs(n_jobs) == 1:
        return df.apply(func, **kwargs)
    else:
        ret = Parallel(n_jobs=n_jobs)(
            delayed(type(df).apply)(df[s], func, **kwargs)
            for s in gen_even_slices(_num_samples(df), effective_n_jobs(n_jobs)))
        return pd.concat(ret)


import os
from typing import Union
from joblib import Parallel, cpu_count, delayed 
import pandas as pd 
import numpy as np 

from lost_ds.functional.split import split_by_empty
from lost_ds.functional.validation import (validate_empty_images, 
                                           validate_img_paths,
                                           validate_single_labels)
from lost_ds.functional.transform import (polygon_to_bbox, to_abs, to_rel,
                                          transform_bbox_style, to_coco)
from lost_ds.detection.bbox_merge import bbox_merge
from lost_ds.io.file_man import FileMan
import fsspec
from lost_ds.util import get_fs


def bbox_nms(df: pd.DataFrame, lbl_col='anno_lbl', method='nms', iou_thr=0.5, 
             filesystem=None, parallel=-1, **kwargs):
    """apply nms to dataframe to suppress overlapping bboxes

    Args:
        df (pd.DataFrame): dataframe containing bboxes to supress
        lbl_col (str, optional): columns containing labels. Defaults to 'anno_lbl'.
        method (str, optional): one of {'nms', 'nmw', 'soft_nms', 'wbf', 'merge'}. Defaults to 'nms'.
        iou_thr (float, optional): iou threshold when to supress bboxes. Defaults to 0.5.
        filesystem (fsspec, optional): filesystem. Defaults to None.

    Returns:
        pd.DataFrame: dataframe with bboxes pruned by given method
    """
    # refer to: https://github.com/ZFTurbo/Weighted-Boxes-Fusion
    from ensemble_boxes import (nms, non_maximum_weighted, soft_nms, 
                                weighted_boxes_fusion)
    df = validate_empty_images(df)
    df, empty_df = split_by_empty(df)
    bbox_df = df[df['anno_dtype']=='bbox']
    bbox_df = validate_single_labels(bbox_df, lbl_col, 'nms_lbl')
    got_scores = False
    if 'anno_confidence' in bbox_df.keys():
        if len(bbox_df['anno_confidence'].notnull()):
            print('Warning! Got data with missing anno_confidence. Will remove it')
        bbox_df = bbox_df[bbox_df['anno_confidence'].notnull()]
        got_scores = True
    else: 
        bbox_df['anno_confidence'] = 1.0
    
    # make same format
    bbox_df = to_rel(bbox_df, filesystem=filesystem, verbose=False)
    bbox_df = transform_bbox_style('x1y1x2y2', bbox_df)
    
    fn_dict = {'nms': nms, 'nmw': non_maximum_weighted, 'soft_nms': soft_nms, 
               'wbf': weighted_boxes_fusion, 'merge': bbox_merge}
    
    # run nms imagewise
    assert method in fn_dict.keys(), f'invalid nms method: {method} - ' \
        f'choose one of {list(fn_dict.keys())}'
    nms_fn = fn_dict[method]
    
    def apply_nms(img_path, img_df):
        boxes_list = np.vstack(img_df['anno_data']).tolist()
        scores_list = list(img_df['anno_confidence'])
        labels_list = list(img_df['nms_lbl'])
        boxes, scores, labels = nms_fn([boxes_list], [scores_list], 
                                       [labels_list], iou_thr=iou_thr, **kwargs)
        n_boxes = len(boxes)
        result = {'anno_data': list(boxes), lbl_col:list(labels), 
                  'anno_style': ['x1y1x2y2']*n_boxes, 
                  'anno_dtype': ['bbox']*n_boxes, 'anno_format': ['rel']*n_boxes, 
                  'img_path': [img_path]*n_boxes}
        if got_scores:
            result['anno_confidence'] = list(scores)
        return pd.DataFrame(result)

    if parallel:
        results = Parallel(parallel)(delayed(apply_nms)(img_path, img_df) 
                            for img_path, img_df in bbox_df.groupby('img_path'))
    else:
        results = list()
        for img_path, img_df in bbox_df.groupby('img_path'):
            results.append(apply_nms(img_path, img_df))
 
    return pd.concat(results + [empty_df])


def coco_eval(gt_df:pd.DataFrame, pred_df:pd.DataFrame, out_dir=None, 
              filesystem: Union[FileMan, fsspec.AbstractFileSystem] = None):
    from pycocotools.coco import COCO
    from pycocotools.cocoeval import COCOeval
    filesystem = get_fs(filesystem)
    if out_dir is None:
        out_dir = '/tmp/coco_eval'
        filesystem.makedirs(out_dir)
    gt_json = os.path.join(out_dir, 'gt.json')
    pred_json = os.path.join(out_dir, 'pred.json')
    
    # to coco
    # predefine img ids
    u_imgs = np.unique(list(gt_df['img_path']) + list(pred_df['img_path']))
    img_mapping = {p: uid for uid, p in enumerate(u_imgs)}
    # predefine lbl ids
    gt_df = validate_single_labels(gt_df, dst_col='coco_lbl')
    pred_df = validate_single_labels(pred_df, dst_col='coco_lbl')
    u_lbls = np.unique(list(gt_df['coco_lbl']) + list(pred_df['coco_lbl']))
    u_lbls = [c for c in u_lbls if pd.notna(c)]
    lbl_mapping = {p: uid+1 for uid, p in enumerate(u_lbls)}
    # dataframe to coco
    coco_gt = to_coco(gt_df, predef_img_mapping=img_mapping, 
                      predef_lbl_mapping=lbl_mapping, json_path=gt_json, 
                      filesystem=filesystem)
    coco_pred = to_coco(pred_df, predef_img_mapping=img_mapping, 
                        predef_lbl_mapping=lbl_mapping, 
                        json_path=pred_json, filesystem=filesystem)
    
    coco_gt = COCO(gt_json)
    coco_pred = COCO(pred_json)
    
    # coco eval
    coco_eval = COCOeval(coco_gt, coco_pred, iouType='bbox')
    coco_eval.evaluate()
    coco_eval.accumulate()
    
    class COCOResults:
        def __init__(self, coco_eval:COCOeval):
            self.coco_eval = coco_eval
            classes = coco_eval.params.catIds
            ious = {'class': [], 'ious': [], 'mean_iou':[]}
            for cl in classes:
                class_ious = dict((k, v) for k,v in coco_eval.ious.items() if k[1]==cl)
                max_ious = [iou.max(axis=0).reshape((-1,1)) for iou in class_ious.values() if len(iou)]
                if len(max_ious):
                    max_ious = np.concatenate(max_ious, 0)
                else:
                    max_ious = np.array([np.nan])
                ious['class'].append(cl)
                ious['ious'].append(list(max_ious.reshape((-1))))
                ious['mean_iou'].append(max_ious.mean())
            self.all_IoUs = np.concatenate(ious['ious'], -1)
            self.mean_IoU = self.all_IoUs.mean()
            self.class_IoUs = pd.DataFrame(ious)
            keys = ['APcoco', 'AP50', 'AP75', 'APcoco_small', 'APcoco_medium', 
                    'APcoco_large', 'AR1', 'AR10', 'AR100', 'AR100_small', 
                    'AR100_medium', 'AR100_large']
            self.coco_results = {k:v for k,v in zip(keys, coco_eval.stats)}
        
        def summarize(self):
            self.coco_eval.summarize()
            print('mIoU per class:')
            print(self.class_IoUs[['class', 'mean_iou']])
            print('\nmIoU all classes: ', self.mean_IoU)
            
    coco_results = COCOResults(coco_eval)
    
    return coco_results
    

def voc_eval(gt_df:pd.DataFrame, pred_df:pd.DataFrame, iou_threshold=0.5,
             APMethod='AllPointsInterpolation'):
    from podm.metrics import BoundingBox, get_pascal_voc_metrics, MethodAveragePrecision
    
    # prepare datasets
    # gt
    gt_df = to_abs(gt_df, verbose=False)
    gt_df = polygon_to_bbox(gt_df, 'x1y1x2y2')
    gt_df = transform_bbox_style('x1y1x2y2', gt_df)
    gt_df = validate_empty_images(gt_df)
    gt_df = validate_single_labels(gt_df, dst_col='anno_lbl')
    # pred
    pred_df = to_abs(pred_df, verbose=False)
    pred_df = polygon_to_bbox(pred_df, 'x1y1x2y2')
    pred_df = transform_bbox_style('x1y1x2y2', pred_df)
    pred_df = validate_empty_images(pred_df)
    pred_df = validate_single_labels(pred_df, dst_col='anno_lbl')
    
    def _cast_df(df):
        boxes = list()
        for path, path_df in df.groupby('img_path'):
            if not path_df['anno_data'].notnull().any():
                continue
            for idx, row in path_df.iterrows():
                bb = BoundingBox.of_bbox(row.img_path, row.anno_lbl,
                                        *list(row.anno_data.flatten()), # ann.xtl, ann.ytl, ann.xbr, ann.ybr, 
                                        row.anno_confidence)
                boxes.append(bb)
        return boxes    
    if not 'anno_confidence' in gt_df.keys():
        gt_df['anno_confidence'] = 1
    gt_bboxes = _cast_df(gt_df)
    pred_bboxes = _cast_df(pred_df)
    if 'allpoints' in APMethod.lower():
        method = MethodAveragePrecision.AllPointsInterpolation
    elif 'elevenpoints' in APMethod.lower():
        method = MethodAveragePrecision.ElevenPointsInterpolation
    
    if not isinstance(iou_threshold, (list, np.ndarray)):
        iou_threshold = [iou_threshold]
    
    data = {'class':[], 'iou_threshold':[], 'ap':[], 'precision':[], 'recall':[], 
            'interpolated_precision':[], 'interpolated_recall':[], 
            'tp':[], 'fp':[], 'fn':[], 'n_gt':[], 'n_det':[]}
    
    for iou_th in iou_threshold:
        results = get_pascal_voc_metrics(gt_bboxes, pred_bboxes, iou_th, method)    
        for cls, metric in results.items():
            data['class'].append(cls)
            data['iou_threshold'].append(float(iou_th))
            data['ap'].append(metric.ap)
            data['precision'].append(metric.precision)
            data['recall'].append(metric.recall)
            data['interpolated_precision'].append(metric.interpolated_precision)
            data['interpolated_recall'].append(metric.interpolated_recall)
            data['tp'].append(metric.tp)
            data['fp'].append(metric.fp)
            data['fn'].append(metric.num_groundtruth - metric.tp)
            data['n_gt'].append(metric.num_groundtruth)
            data['n_det'].append(metric.num_detection)
    
    df = pd.DataFrame(data)
    
    return df


def voc_score_iou_multiplex(gt_df, pred_df, score_key='anno_confidence', 
                            score_range=[0.25, 0.75], score_stepwidth=0.05, 
                            iou_range=[0.25, 0.9], iou_stepwidth=0.05):
    score_shift = np.arange(*score_range, score_stepwidth)
    def shift_score(score_th, gt, pred):
        iou_shift = np.arange(*iou_range, iou_stepwidth)
        pred = pred[pred[score_key] >= score_th]
        voc_df = voc_eval(gt, pred, iou_threshold=iou_shift)
        voc_df['score_threshold'] = score_th
        return voc_df
    # with ThreadPool(cpu_count()) as tp:
    #     evals = tp.map(shift_score, score_shift)
        
    evals = Parallel(-1)(delayed(shift_score)(score_th, gt_df.copy(), pred_df.copy()) 
                         for score_th in score_shift)
    
    voc_df = pd.concat(evals)
    
    return voc_df


def detection_dataset(df, lbl_col='anno_lbl', det_col='det_lbl', 
                      bbox_style='x1y1x2y2', use_empty_images=False, 
                      filesystem=None):
    '''Prepare all bboxes with commonly required operations to use them for 
       detection CNN training
    Args:
        df (pd.DataFrame): Dataframe containing bbox annotations
        lbl_col (str): column name where the anno labels are located (single 
            label or multilabel)
        det_col (str): column name where the training labels are located (single 
            label only)
        bbox_style (str): bbox anno-style. One of {'xywh', 'x1y1x2y2', 'xcycwh'}
        use_empty_images (bool, str, int): specifiy usage of empty images (image 
            without bbox annotation).
            True: keep all images, empty and non-empty
            False: only keep non-empty images and drop all empty images
            'balanced': If more empty images than non-empty ones do exist a 
                random selection will be sampled to have the same amount of 
                empty and non-empty images. If less empty than non-empty images 
                do exist all of them will be kept
            int: a specific amount of empty images will be samples randomly
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
            
    Returns:
        pd.DataFrame: detection dataset

    Note:
        Other anno-types than 'bbox' will be ignored. You can transform 
        polygons to bboxes before by calling LOSTDataset.polygon_to_bbox()
    '''
    fs = get_fs(filesystem)
    df = validate_empty_images(df)
    # df = validate_img_paths(df, False, filesystem=fs)
    df = validate_single_labels(df, lbl_col, det_col)
    non_empty_df, empty_df = split_by_empty(df)
    bbox_df = non_empty_df[non_empty_df['anno_dtype'] == 'bbox']
    bbox_df = to_abs(bbox_df, filesystem=fs, verbose=False)
    bbox_df = transform_bbox_style(dst_style=bbox_style, df=bbox_df)
    empty_df.drop_duplicates(subset=['img_path'], inplace=True)
    
    if use_empty_images:            
        n_empty = -1
        if 'int' in str(type(use_empty_images)):
            n_empty = use_empty_images
        elif use_empty_images=='balanced':
            n_empty = len(bbox_df.img_path.unique())
        if n_empty > 0:
            if len(bbox_df.img_path.unique()) > n_empty:
                empty_df = empty_df.sample(n_empty)
        bbox_df = pd.concat([bbox_df, empty_df])
        
    return bbox_df



import numpy as np
from shapely.geometry import LineString, mapping, MultiLineString
import cv2

from lost_ds.geometry.api import Geometry
from lost_ds.vis.geometries import draw_lines

class Line(Geometry):
    
    def __init__(self):
        super().__init__()
        
    
    def to_shapely(self, data):
        return LineString(data)
    
    
    def segmentation(self, segmentation, color, anno_data, anno_format, 
                     anno_style, line_thickness, **kwargs):
        anno_data = self.to_abs(anno_data, anno_format, segmentation.shape)
        if line_thickness is None:
            return segmentation
        cv2.polylines(segmentation, [anno_data.astype(np.int32)], False, color, 
                      line_thickness)
        return segmentation
    
    
    def crop(self, crop_pos, data, **kwargs):
        intersection, intersects = self._crop_intersection(crop_pos, data)
        if not intersects:
            return intersection
        xmin, ymin, xmax, ymax = crop_pos.bounds
        new_lines = []
        if isinstance(intersection, MultiLineString):
            new_lines = list(intersection.geoms)
        else:
            new_lines = [intersection]
            
        for i, line in enumerate(new_lines):
            new_line = np.array(
                mapping(line)['coordinates']) - [xmin, ymin]
            new_lines[i] = new_line.squeeze()
        return new_lines
    
    
    def validate(self, data):
        return len(data.shape)==2 and len(data)>1
    
    
    def _draw(self, img, data, style, text, color, line_thickness, fontscale, 
              **kwargs):
        if line_thickness is None:
            line_thickness = 2
        return draw_lines(img, data, text, color, line_thickness, fontscale)
    
    


import pandas as pd 
import numpy as np
from joblib import Parallel, delayed, cpu_count

from lost_ds.functional.filter import is_multilabel
from lost_ds.functional.split import split_by_empty
from lost_ds.util import get_fs
from lost_ds.geometry.api import LOSTGeometries

def validate_empty_images(df):
    '''Remove empty entries for images where non-empty entries do exist
    Args:
        df (pd.DataFrame): df to apply on
    Returns:
        pd.DataFrame: dataframe with validated empty images
    '''
    df = df.copy()
    not_empty_df, empty_df = split_by_empty(df)
    not_empty_imgs = not_empty_df['img_path'].unique()
    empty_df = empty_df[~empty_df['img_path'].isin(not_empty_imgs)]
    return pd.concat([not_empty_df, empty_df]).sort_index()


def validate_unique_annos(df):
    '''Remove empty entries for images where non-empty entries do exist
    Args:
        df (pd.DataFrame): df to apply on
    Returns:
        pd.DataFrame: dataframe with validated empty images
    '''
    df = df.copy()
    df['data_hash'] = df.anno_data.map(lambda x: hash(str(x)))
    df = df.drop_duplicates(['img_path', 'data_hash'])
    return df.drop(labels='data_hash', axis='columns')


def validate_geometries(df, remove_invalid=True):
    '''Validate that annotations are unique
    Args:
        df (pd.DataFrame): df to apply on
        lost_geometries (LOSTGeometries): instance to validate the geometries
            If None a new one will be Instaciated
    Returns:
        pd.DataFrame: dataframe with validated empty images
    '''
    df = df.copy()
    geom = LOSTGeometries()
    valid_mask = df[['anno_data', 'anno_dtype']].apply(
        lambda x: geom.validate(*x), axis=1)
    if remove_invalid:
        df = df[valid_mask]
    else:
        if False in valid_mask:
            raise Exception('Found invalid geometries! {}'.
                            format(df[~valid_mask]))
    return df


def validate_img_paths(df, remove_invalid=True, filesystem=None):
    '''Validate that paths do exist
    Args:
        remove_invalid (bool): remove non existing paths from the dataframe
        df (pd.DataFrame): df to apply on
        filesystem (fsspec.filesystem, FileMan): filesystem to use. Use local
            if not initialized
    Returns:
        pd.DataFrame: dataframe with validated image paths
    '''
    df = df.copy()
    fs = get_fs(filesystem)
    images = list(df.img_path.unique())
    def _validate_path(image_path):
        valid_images = []
        invalid_images = []
        exist = fs.exists(image_path)
        if exist:
            valid_images.append(image_path)
        elif not exist:
            invalid_images.append(image_path)
        return valid_images, invalid_images
    
    result = Parallel(-1)(delayed(_validate_path)(p) for p in images)
    valid_images, invalid_images = list(), list()
    for r in result:
        valid_images += r[0]
        invalid_images += r[1]
    
    assert len(valid_images) + len(invalid_images) == len(images)
    if len(invalid_images):
        if remove_invalid:
            df = df[~df['img_path'].isin(invalid_images)]
        else:
            raise Exception('Some images do not exist! {}'.
                            format(invalid_images))
    return df


def validate_single_labels(df, lbl_col='anno_lbl', dst_col='my_lbl'):
    '''Validate single labels. If lbl_col is a multilabel column it will be 
        parsed into single labels at dst_col. 
    
    Args:
        df (pd.DataFrame): Dataframe to validate single labels
        lbl_col (str): label col to validate
        ds_col (str): new column containing single label
        
    Returns:
        pd.DataFrame with column dst_col containing single labels
    '''
    df = df.copy()
    if is_multilabel(df, lbl_col):
        df[lbl_col] = df[lbl_col].map(np.unique)
        n_lbl = df[lbl_col].map(len)
        if n_lbl.max() > 1:
            raise Exception('Found multilabels in {}. \n{}'.
                            format(lbl_col, df[n_lbl > 1][lbl_col]))
        df[dst_col] = df[lbl_col].map(lambda x: x[0] if len(x) else np.nan)
    else:
        df[dst_col] = df[lbl_col]
    return df


from lost_ds.geometry.api import LOSTGeometries

from lost_ds.functional.api import(remove_empty,
                                   split_by_empty,
                                   split_by_img_path,
                                   split_train_test,
                                   split_multilabels,
                                   remap_labels,
                                   remap_img_path,
                                   ignore_labels,
                                   img_selection,
                                   is_multilabel,
                                   label_selection,
                                   polygon_to_bbox,
                                   selection_mask,
                                   to_abs,
                                   to_rel,
                                   transform_bbox_style,
                                   to_coco,
                                   unique_labels,
                                   validate_empty_images,
                                   validate_geometries,
                                   validate_img_paths,
                                   validate_unique_annos,
                                   validate_single_labels)

from lost_ds.cropping.api import (DSCropper,
                                  crop_anno,
                                  crop_img,
                                  crop_dataset,
                                  crop_components)

from lost_ds.copy import (copy_imgs,
                          pack_ds)

from lost_ds.im_util import (get_imagesize,
                             pad_image)

from lost_ds.vis.api import (vis_sample,
                             vis_and_store,
                             vis_semantic_segmentation,
                             draw_polygons,
                             draw_boxes,
                             draw_lines,
                             draw_points,
                             draw_text)

from lost_ds.segmentation.api import (semantic_segmentation, 
                                      segmentation_to_lost)

from lost_ds.detection.api import (detection_dataset, bbox_nms, coco_eval, 
                                   voc_eval, voc_score_iou_multiplex)

from lost_ds.masking.api import mask_dataset 

from lost_ds.util import (get_fs, 
                          to_parquet)

from lost_ds.core import LOSTDataset

from lost_ds.compat import old_lds_to_new_lds


__version__='1.2.0-alpha.1'
